/*!
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

 /*!
 * @module bigtable/family
 */

'use strict';

var createErrorClass = require('create-error-class');
var nodeutil = require('util');
var is = require('is');

/**
 * @private
 */
var FamilyError = createErrorClass('FamilyError', function(name) {
  this.message = 'Column family not found: ' + name + '.';
  this.code = 404;
});

/**
 * @type {module:common/grpcServiceObject}
 * @private
 */
var GrpcServiceObject = require('../common/grpc-service-object.js');

/**
 * Create a Family object to interact with your table column families.
 *
 * @constructor
 * @alias module:bigtable/family
 *
 * @example
 * var gcloud = require('gcloud');
 *
 * var bigtable = gcloud.bigtable({
 *   keyFilename: '/path/to/keyfile.json',
 *   projectId: 'grape-spaceship-123',
 *   cluster: 'gcloud-node',
 *   zone: 'us-central1-b'
 * });
 *
 * var table = bigtable.table('prezzy');
 * var family = table.family('follows');
 */
function Family(table, name) {
  var id = Family.formatName_(table.id, name);

  var methods = {

    /**
     * Create a column family.
     *
     * @param {object=} options - See {module:bigtable/table#createFamily}.
     *
     * @example
     * family.create(function(err, family, apiResponse) {
     *   // The column family was created successfully.
     * });
     */
    create: true,

    /**
     * Delete the column family.
     *
     * @param {function=} callback - The callback function.
     * @param {?error} callback.err - An error returned while making this
     *     request.
     * @param {object} callback.apiResponse - The full API response.
     *
     * @example
     * family.delete(function(err, apiResponse) {});
     */
    delete: {
      protoOpts: {
        service: 'BigtableTableService',
        method: 'deleteColumnFamily'
      },
      reqOpts: {
        name: id
      }
    },

    /**
     * Check if the column family exists.
     *
     * @param {function} callback - The callback function.
     * @param {?error} callback.err - An error returned while making this
     *     request.
     * @param {boolean} callback.exists - Whether the family exists or not.
     *
     * @example
     * family.exists(function(err, exists) {});
     */
    exists: true,

    /**
     * Get a column family if it exists.
     *
     * You may optionally use this to "get or create" an object by providing an
     * object with `autoCreate` set to `true`. Any extra configuration that is
     * normally required for the `create` method must be contained within this
     * object as well.
     *
     * @param {options=} options - Configuration object.
     * @param {boolean} options.autoCreate - Automatically create the object if
     *     it does not exist. Default: `false`
     *
     * @example
     * family.get(function(err, family, apiResponse) {
     *   // `family.metadata` has been populated.
     * });
     */
    get: true
  };

  var config = {
    parent: table,
    id: id,
    methods: methods,
    createMethod: function(_, options, callback) {
      table.createFamily(name, options, callback);
    }
  };

  GrpcServiceObject.call(this, config);
}

nodeutil.inherits(Family, GrpcServiceObject);

/**
 * Format the Column Family name into the expected proto format.
 *
 * @private
 *
 * @param {string} tableName - The full formatted table name.
 * @param {string} name - The column family name.
 * @return {string}
 *
 * @example
 * Family.formatName_(
 *   'projects/p/zones/z/clusters/c/tables/t',
 *   'my-family'
 * );
 * // 'projects/p/zones/z/clusters/c/tables/t/columnFamilies/my-family'
 */
Family.formatName_ = function(tableName, name) {
  if (name.indexOf('/') > -1) {
    return name;
  }

  return tableName + '/columnFamilies/' + name;
};

/**
 * Formats Garbage Collection rule into proto format.
 *
 * @private
 *
 * @param {object} ruleObj - The rule object.
 * @return {object}
 *
 * @example
 * Family.formatRule({
 *   age: {
 *     seconds: 10000,
 *     nanos: 10000
 *   },
 *   versions: 2,
 *   union: true
 * });
 * // {
 * //   union: {
 * //     rules: [
 * //       {
 * //         maxAge: {
 * //           seconds: 10000,
 * //           nanos: 10000
 * //         }
 * //       }, {
 * //         maxNumVersions: 2
 * //       }
 * //     ]
 * //   }
 * // }
 */
Family.formatRule_ = function(ruleObj) {
  var rules = [];

  if (ruleObj.age) {
    rules.push({
      maxAge: ruleObj.age
    });
  }

  if (ruleObj.versions) {
    rules.push({
      maxNumVersions: ruleObj.versions
    });
  }

  if (!ruleObj.intersection && !ruleObj.union) {
    return rules[0];
  }

  var rule = {};
  var ruleType = ruleObj.union ? 'union' : 'intersection';

  rule[ruleType] = {
    rules: rules
  };

  return rule;
};

/**
 * Get the column family's metadata.
 *
 * @param {function} callback - The callback function.
 * @param {?error} callback.err - An error returned while making this
 *     request.
 * @param {object} callback.metadata - The metadata.
 * @param {object} callback.apiResponse - The full API response.
 *
 * @example
 * family.getMetadata(function(err, metadata, apiResponse) {});
 */
Family.prototype.getMetadata = function(callback) {
  var self = this;

  this.parent.getFamilies(function(err, families, resp) {
    if (err) {
      callback(err, null, resp);
      return;
    }

    for (var i = 0, l = families.length; i < l; i++) {
      if (families[i].id === self.id) {
        self.metadata = families[i].metadata;
        callback(null, self.metadata, resp);
        return;
      }
    }

    var error = new FamilyError(self.id);
    callback(error, null, resp);
  });
};

/**
 * Set the column family's metadata.
 *
 * See {module:bigtable/table#createFamily} for a detailed explanation of the
 * arguments.
 *
 * @resource [Garbage Collection Proto Docs]{@link https://github.com/googleapis/googleapis/blob/3592a7339da5a31a3565870989beb86e9235476e/google/bigtable/admin/table/v1/bigtable_table_data.proto#L59}
 *
 * @param {object} metadata - Metadata object.
 * @param {object|string=} metadata.rule - Garbage collection rule.
 * @param {string=} metadata.name - The updated column family name.
 * @param {function} callback - The callback function.
 * @param {?error} callback.err - An error returned while making this
 *     request.
 * @param {object} callback.apiResponse - The full API response.
 *
 * @example
 * family.setMetadata({
 *   name: 'updated-name',
 *   rule: 'version() > 3 || (age() > 3d && version() > 1)'
 * }, function(err, apiResponse) {});
 */
Family.prototype.setMetadata = function(metadata, callback) {
  var grpcOpts = {
    service: 'BigtableTableService',
    method: 'updateColumnFamily'
  };

  var reqOpts = {
    name: this.id
  };

  if (metadata.rule) {
    if (is.string(metadata.rule)) {
      reqOpts.gcExpression = metadata.rule;
    } else if (is.object(metadata.rule)) {
      reqOpts.gcRule = Family.formatRule_(metadata.rule);
    }
  }

  if (metadata.name) {
    reqOpts.name = Family.formatName_(this.parent.id, metadata.name);
  }

  this.request(grpcOpts, reqOpts, callback);
};

module.exports = Family;
module.exports.FamilyError = FamilyError;
